# frozen_string_literal: true

# The Puppet Plugins module defines extension points where plugins can be configured
# to add or modify puppet's behavior. See the respective classes in this module for more
# details.
#
# @api public
# @since Puppet 4.0.0
#
module Puppet::Plugins
end
